import java.io.*;
import java.util.*;

/**
 * Class that reads tokens from a simplified font file.
 * 
 * @author Donald Chinn
 * @version April 6, 2003
 */
public class FontFileTokenizer extends StreamTokenizer {

    // used for the result of nextGlyph()
    public int glyphNumber;
    public int glyphData;
    
    // used for the result of nextHash()
    public int hashValue;
    
    // hashValueSentinel indicates the beginning of the hash value
    // section of the font file.
    public static final char hashValueChar = '*';
    public static final String hashValueSentinel = "***";

    // noMoreGlyphs is used in nextGlyph() if there are no more glyphs     
    public static final int noMoreGlyphs = -1;
    
    /*
     * Once the method advanceToData has been called, then
     * it has no effect if called again.
     * Similarly, once advanceToHashValues has been called, then
     * it has no effect if called again.
     */
    private boolean advancedToData;
    private boolean advancedToHashValues;
    
    public FontFileTokenizer(Reader r) {
        super(r);
        this.eolIsSignificant(false);
        this.parseNumbers();
        this.wordChars (hashValueChar, hashValueChar);
        
        advancedToData = false;
        advancedToHashValues = false;
    }
    

    /**
     * Return the number of glyphs in the font file.
     * This method assumes that the file pointer is pointing to the
     * beginning of the font file.
     * An exception is thrown if it can be deduced that the file is
     * not of the right format or the file pointer is beyond the
     * data of the first glyph.
     * @throws  error if an exception occurs or the font file is bad
     */
    public int getNumGlyphs () throws Error {
        int tokenType;
        try {
            tokenType = nextToken();
        }
        catch (Exception e) {
            throw new Error (e.getMessage());
        }
        
        if (tokenType == TT_NUMBER) {
            if (!isValidNumber(nval)) {
                throw new Error ("Bad font file format.");
            } else {
                return (int) nval;
            }
        } else {
            throw new Error ("Bad font file format.");
        }
            
    }


    /**
     * Advance the file pointer to the beginning of the glyph data.
     * This method assumes that the file pointer is pointing to the
     * beginning of the font file.
     * An exception is thrown if it can be deduced that the file is
     * not of the right format or the file pointer is not at the
     * beginning of the font file.
     * @throws  error if an exception occurs or the font file is bad
     */
    public void advanceToData () throws Error {
        if (!advancedToData) {
            int tokenType;
            try {
                tokenType = nextToken();
            }
            catch (Exception e) {
                throw new Error (e.getMessage());
            }
            
            if (tokenType == TT_NUMBER) {
                if (!isValidNumber(nval)) {
                    throw new Error ("Bad font file format.");
                }
                // assume that we have read the number of glyphs.  Stop.
            } else {
                throw new Error ("Bad font file format.");
            }
            
            advancedToData = true;
        }
    }
    
    
    /**
     * Advance the file pointer to the beginning of the hash
     * values.  This method assumes that the file pointer is pointing
     * to somewhere in the file before the hash values.
     * @throws  error if an exception occurs or the font file is bad
     */
    public void advanceToHashValues () throws Error {
        if (!advancedToHashValues) {

            boolean done = false;   // have we found the beginning of the hash values?
            while (!done) {
                int tokenType;
                try {
                    tokenType = nextToken();
                }
                catch (Exception e) {
                    throw new Error (e.getMessage());
                }
                switch (tokenType) {
                    case TT_WORD:
                        if (sval.equals(hashValueSentinel)) {
                            done = true;
                        } else {
                            throw new Error ("Bad font file format.");
                        }
                        // System.out.println(sval);
                        break;
                    case TT_NUMBER:
                        if (!isValidNumber(nval)) {
                            throw new Error ("Bad font file format.");
                        }
                        // System.out.println(nval);
                        break;
                    case TT_EOF:
                        throw new Error ("Could not find hash values.");
                }
            }
            
            advancedToData = true;
            advancedToHashValues = true;
        }
    }


    /**
     * Store in glyphNumber and glyphData the next (glyphNumber, data) pair.
     * This method assumes that the file pointer is pointing to
     * a (glyphNumber, data) pair.  If there are no more glyphs, then
     * glyphNumber and glyphData are set to (noMoreGlyphs, noMoreGlyphs).
     * 
     * @throws error if an IOException occurs or the font file is bad
     */
    public void nextGlyph() throws Error {
        int tokenType;
        try {
            tokenType = nextToken();
        }
        catch (IOException e) {
            throw new Error (e.getMessage());
        }
        if ((tokenType == TT_NUMBER) && isValidNumber(nval)) {
            glyphNumber = (int) nval;
        } else if (((tokenType == TT_WORD) && sval.equals(hashValueSentinel)) ||
                   (tokenType == TT_EOF)) {
            // no more glyphs
            glyphNumber = noMoreGlyphs;
            glyphData = noMoreGlyphs;
            return;
        } else {
            throw new Error ("Bad glyph number.");
        }
        
        try {
            tokenType = nextToken();
        }
        catch (IOException e) {
            throw new Error (e.getMessage());
        }
        if ((tokenType == TT_NUMBER) && isValidNumber(nval)) {
            glyphData = (int) nval;
        } else {
            throw new Error ("Bad glyph data.");
        }
    }

    
    /**
     * Store in hashValue the next hash value.
     * This method assumes that the file pointer is pointing to
     * a hash value.
     * 
     * @throws error if an IOException occurs or the font file is bad
     */
    public void nextHash() throws Error, EOFException {
        int tokenType;
        try {
            tokenType = nextToken();
        }
        catch (IOException e) {
            throw new Error (e.getMessage());
        }
        if ((tokenType == TT_NUMBER) && isValidNumber(nval)) {
            hashValue = (int) nval;
        } else if (tokenType == TT_EOF) {
            throw new EOFException ();
        } else {
            throw new Error ("Bad hash value.");
        }
    }

    
    /**
     * Check if the number is a valid number.  A valid number is
     * an integer between 0 and 2^16 - 1  (inclusive).
     * @param value  number to be checked
     * @return true, if the value is valid; false otherwise
     */
    private boolean isValidNumber(double value) {
        return ((value == Math.floor(value)) &&
                (value >= 0) &&
                (value <= 65535));
    }
}
